# -*- coding: utf-8 -*-
#
# Copyright 2009-2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the main SSO client code."""

import logging

from twisted.internet import defer
from ubuntuone.devtools.handlers import MementoHandler

from ubuntu_sso import main
from ubuntu_sso.main import (
    CredentialsManagement,
    except_to_errdict,
    SSOLogin,
    thread_execute,
    TIMEOUT_INTERVAL,
    UbuntuSSOService,
)
from ubuntu_sso.main import (HELP_TEXT_KEY, PING_URL_KEY,
    TC_URL_KEY, UI_CLASS_KEY, UI_MODULE_KEY, WINDOW_ID_KEY,
    SUCCESS_CB_KEY, ERROR_CB_KEY, DENIAL_CB_KEY)
from ubuntu_sso.main.tests import FakedCredentials
from ubuntu_sso.tests import (APP_NAME, TC_URL, HELP_TEXT, CAPTCHA_ID,
    CAPTCHA_SOLUTION, EMAIL, EMAIL_TOKEN, NAME, PASSWORD, PING_URL, TOKEN,
    WINDOW_ID, Recorder, TestCase)


# Access to a protected member 'yyy' of a client class
# pylint: disable=W0212


class SampleException(Exception):
    """The exception that will be thrown by the fake thread_execute."""


def fake_ok_thread_execute(f, app, cb, eb):
    """A fake thread_execute function that succeeds."""
    cb(app, f())


def fake_err_thread_execute(f, app, cb, eb):
    """A fake thread_execute function that fails."""
    try:
        f()
    except Exception, e:  # pylint: disable=W0703
        eb(app, except_to_errdict(e))
    else:
        eb(app, except_to_errdict(SampleException()))


class FakedProxy(Recorder):
    """A faked multiplatform proxy."""


class FakedAccount(Recorder):
    """A faked Account processor."""


class FakedCredentialsFactory(Recorder):
    """A very dummy Credentials object."""

    credentials = FakedCredentials()

    # pylint: disable=C0103

    def Credentials(self, *a, **kw):
        """Return always the same Credentials instance."""
        return self.credentials

    # pylint: enable=C0103


class TestExceptToErrdictException(Exception):
    """A dummy exception for the following testcase."""


class ExceptToErrdictTestCase(TestCase):
    """Tests for the except_to_errdict function."""

    def test_first_arg_is_dict(self):
        """If the first arg is a dict, use it as the base dict."""
        sample_dict = {
            "errorcode1": "error message 1",
            "errorcode2": "error message 2",
            "errorcode3": "error message 3",
        }
        e = TestExceptToErrdictException(sample_dict)
        result = except_to_errdict(e)

        self.assertEqual(result["errtype"], e.__class__.__name__)
        for k in sample_dict.keys():
            self.assertIn(k, result)
            self.assertEqual(result[k], sample_dict[k])

    def test_first_arg_is_str(self):
        """If the first arg is a str, use it as the message."""
        sample_string = "a sample string"
        e = TestExceptToErrdictException(sample_string)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], sample_string)

    def test_first_arg_is_unicode(self):
        """If the first arg is a unicode, use it as the message."""
        sample_string = u"a sample string"
        e = TestExceptToErrdictException(sample_string)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], sample_string)

    def test_no_args_at_all(self):
        """If there are no args, use the class docstring."""
        e = TestExceptToErrdictException()
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)
        self.assertEqual(result["message"], e.__class__.__doc__)

    def test_some_other_thing_as_first_arg(self):
        """If first arg is not basestring nor dict, then repr all args."""
        sample_args = (None, u"unicode2\ufffd", "errorcode3")
        e = TestExceptToErrdictException(*sample_args)
        result = except_to_errdict(e)
        self.assertEqual(result["errtype"], e.__class__.__name__)


class BaseTestCase(TestCase):
    """Base test case."""

    timeout = 2

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        self.proxy = FakedProxy()

    def assert_recorder_called(self, fake, method, *args, **kwargs):
        """Check that 'fake.method(*args, **kwargs)' was called."""
        self.assertEqual(fake._called[method], [(args, kwargs)])


class SSOLoginTestCase(BaseTestCase):
    """Test the SSOLogin class."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SSOLoginTestCase, self).setUp()

        def ksc(keyring, k, val):
            """Assert over token and app_name."""
            self.assertEqual(k, APP_NAME)
            self.assertEqual(val, TOKEN)
            self.keyring_was_set = True
            self.keyring_values = k, val
            return defer.succeed(None)

        self.patch(main.Keyring, "set_credentials", ksc)
        self.patch(main, 'Account', FakedAccount)
        self.keyring_was_set = False
        self.keyring_values = None

        self.obj = SSOLogin(self.proxy)

    def test_creation(self):
        """Test that the object creation is successful."""
        self.assertIsInstance(self.obj.processor, main.Account)
        self.assertTrue(self.obj.proxy is self.proxy)

    @defer.inlineCallbacks
    def test_generate_captcha(self):
        """Test that the captcha method works ok."""
        d = defer.Deferred()
        filename = "sample filename"
        expected_result = "expected result"

        self.patch(self.obj, "CaptchaGenerated", lambda *a: d.callback(a))
        self.patch(self.obj, "CaptchaGenerationError", d.errback)

        self.obj.processor._next_result = defer.succeed(expected_result)
        self.obj.generate_captcha(APP_NAME, filename)

        app_name, result = yield d
        self.assertEqual(result, expected_result)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor, 'generate_captcha',
            filename)

    @defer.inlineCallbacks
    def test_register_user(self):
        """Test that the register_user method works ok."""
        d = defer.Deferred()
        expected_result = "expected result"

        self.patch(self.obj, "UserRegistered", lambda *a: d.callback(a))
        self.patch(self.obj, "UserRegistrationError", d.errback)

        self.obj.processor._next_result = defer.succeed(expected_result)
        self.obj.register_user(APP_NAME, EMAIL, PASSWORD, NAME, CAPTCHA_ID,
                               CAPTCHA_SOLUTION)

        app_name, result = yield d
        self.assertEqual(result, expected_result)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor, 'register_user',
            EMAIL, PASSWORD, NAME, CAPTCHA_ID, CAPTCHA_SOLUTION)

    @defer.inlineCallbacks
    def test_login(self):
        """Test that the login method works ok."""
        d = defer.Deferred()

        self.patch(self.obj, "LoggedIn", lambda *a: d.callback(a))
        self.patch(self.obj, "LoginError", d.errback)
        self.patch(self.obj, "UserNotValidated", d.errback)

        self.obj.processor._next_result = TOKEN
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, result = yield d
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(self.keyring_was_set, "The keyring should be set")
        self.assert_recorder_called(self.obj.processor, 'login',
            EMAIL, PASSWORD, main.get_token_name(APP_NAME))

    @defer.inlineCallbacks
    def test_validate_email(self):
        """Test that the validate_email method works ok."""
        d = defer.Deferred()

        self.patch(self.obj, "EmailValidated", lambda *a: d.callback(a))
        self.patch(self.obj, "EmailValidationError", d.errback)

        self.obj.processor._next_result = TOKEN
        self.obj.validate_email(APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN)

        app_name, result = yield d
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(self.keyring_was_set, "The keyring should be set")
        self.assert_recorder_called(self.obj.processor, 'validate_email',
            EMAIL, PASSWORD, EMAIL_TOKEN, main.get_token_name(APP_NAME))

    @defer.inlineCallbacks
    def test_request_password_reset_token(self):
        """Test that the request_password_reset_token method works ok."""
        d = defer.Deferred()

        self.patch(self.obj, "PasswordResetTokenSent",
                   lambda *a: d.callback(a))
        self.patch(self.obj, "PasswordResetError", d.errback)

        self.obj.processor._next_result = EMAIL
        self.obj.request_password_reset_token(APP_NAME, EMAIL)

        app_name, result = yield d
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor,
            'request_password_reset_token', EMAIL)

    @defer.inlineCallbacks
    def test_set_new_password(self):
        """Test that the set_new_password method works ok."""
        d = defer.Deferred()

        self.patch(self.obj, "PasswordChanged", lambda *a: d.callback(a))
        self.patch(self.obj, "PasswordChangeError", d.errback)

        self.obj.processor._next_result = EMAIL
        self.obj.set_new_password(APP_NAME, EMAIL, EMAIL_TOKEN, PASSWORD)

        app_name, result = yield d
        self.assertEqual(result, EMAIL)
        self.assertEqual(app_name, APP_NAME)
        self.assert_recorder_called(self.obj.processor,
            'set_new_password', EMAIL, EMAIL_TOKEN, PASSWORD)


class SSOLoginWithErrorTestCase(SSOLoginTestCase):
    """Test the SSOLogin class."""

    def sample_fail(self, *args):
        """Fail predictably."""
        raise SampleException("Sample message")

    @defer.inlineCallbacks
    def test_generate_captcha(self):
        """Test that the captcha method fails as expected."""
        d = defer.Deferred()
        filename = "sample filename"

        self.patch(self.obj, "CaptchaGenerated", d.errback)
        self.patch(self.obj, "CaptchaGenerationError",
                   lambda *a: d.callback(a))
        self.patch(self.obj.processor, "generate_captcha", self.sample_fail)
        self.obj.generate_captcha(APP_NAME, filename)

        app_name, error = yield d
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_register_user(self):
        """Test that the register_user method fails as expected."""
        d = defer.Deferred()

        self.patch(self.obj, "UserRegistered", d.errback)
        self.patch(self.obj, "UserRegistrationError", lambda *a: d.callback(a))
        self.patch(self.obj.processor, "register_user", self.sample_fail)
        self.obj.register_user(APP_NAME, EMAIL, PASSWORD, NAME, CAPTCHA_ID,
                             CAPTCHA_SOLUTION)

        app_name, error = yield d
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_login(self):
        """The login method fails as expected when get_token_name fails."""
        d = defer.Deferred()

        def fake_gtn(*args):
            """A fake get_token_name that fails."""
            raise SampleException()

        self.patch(main, "get_token_name", fake_gtn)
        self.patch(self.obj, "LoggedIn", d.errback)
        self.patch(self.obj, "LoginError", lambda *a: d.callback(a))
        self.patch(self.obj, "UserNotValidated", d.errback)
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, error = yield d
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_login_set_credentials(self):
        """The login method fails as expected when set_credentials fails."""
        d = defer.Deferred()

        def fake_set_creds(*args):
            """A fake Keyring.set_credentials that fails."""
            return defer.fail(SampleException())

        self.patch(main.Keyring, "set_credentials", fake_set_creds)
        self.patch(self.obj.processor, "login",
                                        lambda *args: defer.succeed(None))
        self.patch(self.obj.processor, "is_validated",
                                        lambda *args: defer.succeed(True))
        fail = lambda app, res: d.errback((app, res))
        self.patch(self.obj, "LoggedIn", fail)
        self.patch(self.obj, "LoginError", lambda *a: d.callback(a))
        self.patch(self.obj, "UserNotValidated", fail)
        self.obj.login(APP_NAME, EMAIL, PASSWORD)

        app_name, error = yield d
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_validate_email(self):
        """Test that the validate_email method fails as expected."""
        d = defer.Deferred()

        def fake_gtn(*args):
            """A fake get_token_name that fails."""
            raise SampleException()

        self.patch(main, "get_token_name", fake_gtn)

        self.patch(self.obj, "EmailValidated", d.errback)
        self.patch(self.obj, "EmailValidationError", lambda *a: d.callback(a))
        self.obj.validate_email(APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN)

        app_name, error = yield d
        self.assertEqual(app_name, APP_NAME)
        self.assertTrue(isinstance(error, SampleException))
        self.assertFalse(self.keyring_was_set, "Keyring should not be set")

    @defer.inlineCallbacks
    def test_request_password_reset_token(self):
        """Test the request_password_reset_token method fails as expected."""
        d = defer.Deferred()

        self.patch(self.obj.processor, "request_password_reset_token",
                   lambda _: defer.fail(SampleException()))
        self.patch(self.obj, "PasswordResetTokenSent", d.errback)
        self.patch(self.obj, "PasswordResetError", lambda *a: d.callback(a))
        self.obj.request_password_reset_token(APP_NAME, EMAIL)

        app_name, error = yield d
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_set_new_password(self):
        """Test that the set_new_password method fails as expected."""
        d = defer.Deferred()

        self.patch(self.obj.processor, "set_new_password",
                   lambda *args: defer.fail(SampleException()))
        self.patch(self.obj, "PasswordChanged", d.errback)
        self.patch(self.obj, "PasswordChangeError", lambda *a: d.callback(a))
        self.obj.set_new_password(APP_NAME, EMAIL, EMAIL_TOKEN, PASSWORD)

        app_name, error = yield d
        self.assertTrue(isinstance(error, SampleException))
        self.assertEqual(app_name, APP_NAME)


class BlockingFunctionTestCase(TestCase):
    """Tests for the "thread_execute" function."""

    timeout = 5

    @defer.inlineCallbacks
    def test_thread_execute(self):
        """Test the normal behaviour."""
        d = defer.Deferred()
        expected_result = "expected result"

        def f():
            """No failure."""
            return expected_result

        thread_execute(f, APP_NAME, lambda *a: d.callback(a), d.errback)

        app_name, result = yield d

        self.assertEqual(result, expected_result)
        self.assertEqual(app_name, APP_NAME)

    @defer.inlineCallbacks
    def test_thread_execute_error(self):
        """Test the behaviour when an Exception is raised."""
        d = defer.Deferred()
        expected_error_message = "expected error message"
        expected_exc = SampleException(expected_error_message)

        def f():
            """Failure."""
            raise expected_exc

        thread_execute(f, APP_NAME, d.errback, lambda *a: d.callback(a))

        app_name, error = yield d

        self.assertEqual(app_name, APP_NAME)
        self.assertEqual(error, expected_exc)


class CredentialsManagementTestCase(BaseTestCase):
    """Tests for the CredentialsManagement DBus interface."""

    base_args = {
        HELP_TEXT_KEY: HELP_TEXT, PING_URL_KEY: PING_URL,
        TC_URL_KEY: TC_URL, WINDOW_ID_KEY: WINDOW_ID,
        UI_CLASS_KEY: 'SuperUI', UI_MODULE_KEY: 'foo.bar.baz',
    }

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementTestCase, self).setUp()

        self.factory = FakedCredentialsFactory()
        self.patch(main, 'Credentials', self.factory.Credentials)
        self.obj = CredentialsManagement(timeout_func=lambda *a: None,
                                         shutdown_func=lambda *a: None,
                                         proxy=self.proxy)
        self.args = {}
        self.cred_args = {}

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        main.logger.addHandler(self.memento)
        self.addCleanup(main.logger.removeHandler, self.memento)


class CredentialsManagementRefCountingTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement ref counting."""

    @defer.inlineCallbacks
    def assert_refcounter_increased(self, method_name, signal_name=None):
        """Assert that calling 'method_name' increases the ref counter."""
        if signal_name is not None:
            d = defer.Deferred()
            self.patch(self.obj, signal_name, lambda *a: d.callback(a))
        else:
            d = defer.succeed(None)

        getattr(self.obj, method_name)(APP_NAME, self.args)

        yield d
        self.assertEqual(self.obj.ref_count, 1)

    def assert_refcounter_decreased(self, signal_name, *a, **kw):
        """Assert that calling 'signal_name' decreases the ref counter."""
        self.obj.ref_count = 3

        getattr(self.obj, signal_name)(*a, **kw)

        self.assertEqual(self.obj.ref_count, 2)

    def assert_refcounter_negative(self, signal_name, *a, **kw):
        """Assert that calling 'signal_name' decreases the ref counter.

        If decreased value is negative, assert that a warning log was made.

        """
        self.obj._ref_count = -3  # overwrite internal to force a negative

        getattr(self.obj, signal_name)(*a, **kw)

        self.assertEqual(self.obj.ref_count, 0)
        msg = 'Attempting to decrease ref_count to a negative value (-4).'
        self.assertTrue(self.memento.check_warning(msg))

    def test_ref_counting(self):
        """Ref counting is in place."""
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_find_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('find_credentials',
                                               'CredentialsFound')

    @defer.inlineCallbacks
    def test_find_credentials_with_success_cb(self):
        """Keep proper track of ongoing requests."""
        d = defer.Deferred()
        counter = []

        def in_the_middle():
            """Store the ref_count value."""
            counter.append(self.obj.ref_count)
            return defer.succeed(TOKEN)

        self.patch(self.factory.credentials, 'find_credentials', in_the_middle)
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.callback, error_cb=d.errback)
        yield d
        self.assertEqual(counter, [1])
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_find_credentials_with_error_cb(self):
        """Keep proper track of ongoing requests."""
        d = defer.Deferred()
        counter = []

        def in_the_middle():
            """Store the ref_count value."""
            counter.append(self.obj.ref_count)
            return defer.fail('foo')

        self.patch(self.factory.credentials, 'find_credentials', in_the_middle)
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.errback, error_cb=d.callback)

        yield d
        self.assertEqual(counter, [1])
        self.assertEqual(self.obj.ref_count, 0)

    @defer.inlineCallbacks
    def test_clear_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('clear_credentials',
                                               'CredentialsCleared')

    @defer.inlineCallbacks
    def test_store_credentials(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('store_credentials',
                                               'CredentialsStored')

    def test_register(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('register')

    def test_login(self):
        """Keep proper track of ongoing requests."""
        yield self.assert_refcounter_increased('login')

    def test_several_requests(self):
        """Requests can be nested."""
        self.obj.login(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)
        self.obj.login(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)
        self.obj.register(APP_NAME, self.args)

        self.assertEqual(self.obj.ref_count, 5)

    def test_credentials_found(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsFound', APP_NAME, TOKEN)

    def test_credentials_not_found(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsNotFound', APP_NAME)

    def test_credentials_cleared(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsCleared', APP_NAME)

    def test_credentials_stored(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsStored', APP_NAME)

    def test_credentials_error(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('CredentialsError', APP_NAME,
                                         SampleException('test'))

    def test_authorization_denied(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_decreased('AuthorizationDenied', APP_NAME)

    def test_credentials_found_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsFound', APP_NAME, TOKEN)

    def test_credentials_not_found_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsNotFound', APP_NAME)

    def test_credentials_cleared_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsCleared', APP_NAME)

    def test_credentials_stored_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsStored', APP_NAME)

    def test_credentials_error_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('CredentialsError', APP_NAME,
                                        SampleException('test'))

    def test_autorization_denied_when_ref_count_is_not_positive(self):
        """Ref counter is decreased when a signal is sent."""
        self.assert_refcounter_negative('AuthorizationDenied', APP_NAME)

    def test_on_zero_ref_count_shutdown(self):
        """When ref count reaches 0, queue shutdown op."""
        self.patch(self.obj, 'timeout_func', self._set_called)
        self.obj.login(APP_NAME, self.args)
        self.obj.CredentialsFound(APP_NAME, TOKEN)

        self.assertEqual(self._called,
                         ((TIMEOUT_INTERVAL, self.obj.shutdown), {}))

    def test_on_non_zero_ref_count_do_not_shutdown(self):
        """If ref count is not 0, do not queue shutdown op."""
        self.patch(self.obj, 'timeout_func', self._set_called)
        self.obj.login(APP_NAME, self.args)

        self.assertEqual(self._called, False)

    def test_on_non_zero_ref_count_after_zero_do_not_shutdown(self):
        """If the shutdown was queued, do not quit if counter is not zero."""

        def fake_timeout_func(interval, func):
            """Start a new request when the timer is started."""
            self.obj.register(APP_NAME, self.args)
            assert self.obj.ref_count > 0
            func()

        self.patch(self.obj, 'timeout_func', fake_timeout_func)
        self.patch(self.obj, 'shutdown_func', self._set_called)

        self.obj.login(APP_NAME, self.args)
        self.obj.CredentialsFound(APP_NAME, TOKEN)
        # counter reached 0, timeout_func was called

        self.assertEqual(self._called, False, 'shutdown_func was not called')

    def test_zero_ref_count_after_zero_do_shutdown(self):
        """If the shutdown was queued, do quit if counter is zero."""

        def fake_timeout_func(interval, func):
            """Start a new request when the timer is started."""
            assert self.obj.ref_count == 0
            func()

        self.patch(self.obj, 'timeout_func', fake_timeout_func)
        self.patch(self.obj, 'shutdown_func', self._set_called)

        self.obj.login(APP_NAME, self.args)
        self.obj.CredentialsFound(APP_NAME, TOKEN)
        # counter reached 0, timeout_func was called

        self.assertEqual(self._called, ((), {}), 'shutdown_func was called')


class CredentialsManagementClearTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement clear method."""

    method = 'clear_credentials'
    success_signal = 'CredentialsCleared'
    success_result = (APP_NAME,)
    others_should_errback = []

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementClearTestCase, self).setUp()
        self.call_method = getattr(self.obj, self.method)

    def test_backend_called(self):
        """The credentials backend is properly called."""
        self.call_method(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)

    @defer.inlineCallbacks
    def test_does_not_block(self):
        """Calling 'method' does not block but return thru signals."""
        d = defer.Deferred()

        self.patch(self.obj, self.success_signal, lambda *a: d.callback(a))
        self.patch(self.obj, 'CredentialsError',
                   lambda *a: d.errback(SampleException(a)))
        for signal in self.others_should_errback:
            self.patch(self.obj, signal,
                       lambda *a: d.errback(AssertionError(a)))

        self.call_method(APP_NAME, self.args)

        result = yield d
        self.assertEqual(result, self.success_result)

    @defer.inlineCallbacks
    def test_handles_error(self):
        """If calling the backend fails, CredentialsError is sent."""
        d = defer.Deferred()

        self.patch(self.obj, self.success_signal, d.errback)
        self.patch(self.obj, 'CredentialsError', lambda *a: d.callback(a))

        exc = SampleException('baz')
        self.patch(self.factory.credentials, self.method,
                   lambda *a: defer.fail(exc))
        self.call_method(APP_NAME, self.args)

        app_name, error = yield d
        self.assertEqual(error, exc)
        self.assertEqual(app_name, APP_NAME)


class CredentialsManagementStoreTestCase(CredentialsManagementClearTestCase):
    """Tests for the CredentialsManagement store method."""

    method = 'store_credentials'
    success_signal = 'CredentialsStored'


class CredentialsManagementFindTestCase(CredentialsManagementClearTestCase):
    """Tests for the CredentialsManagement find method."""

    method = 'find_credentials'
    success_signal = 'CredentialsFound'
    success_result = (APP_NAME, TOKEN)
    others_should_errback = ['CredentialsNotFound']

    @defer.inlineCallbacks
    def test_find_credentials_with_success_cb(self):
        """The credentials are asked and returned in a thread_execute call."""
        d = defer.Deferred()

        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.callback, error_cb=d.errback)

        creds = yield d
        expected = yield self.factory.credentials.find_credentials()
        self.assertEqual(creds, expected)
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)

    @defer.inlineCallbacks
    def test_find_credentials_with_error_cb(self):
        """If find_credentials fails, error_cb is called."""
        d = defer.Deferred()

        self.patch(self.factory.credentials, 'find_credentials',
                   lambda *a: defer.fail(SampleException('baz')))
        self.obj.find_credentials(APP_NAME, self.args,
                                  success_cb=d.errback, error_cb=d.callback)

        errdict = yield d
        self.assertEqual(errdict["errtype"], "SampleException")
        self.assert_recorder_called(self.factory, 'Credentials', APP_NAME)


class CredentialsManagementNotFindTestCase(CredentialsManagementFindTestCase):
    """Tests for the CredentialsManagement find method."""

    success_signal = 'CredentialsNotFound'
    success_result = (APP_NAME,)
    others_should_errback = ['CredentialsFound']

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementNotFindTestCase, self).setUp()
        self.call_method = getattr(self.obj, self.method)
        self.patch(self.factory.credentials, self.method,
                   lambda: defer.succeed({}))


class CredentialsManagementOpsTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement login/register methods."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementOpsTestCase, self).setUp()
        self.args = dict((k, str(v)) for k, v in self.base_args.iteritems())
        self.cred_args = self.base_args.copy()
        self.cred_args[SUCCESS_CB_KEY] = self.obj.CredentialsFound
        self.cred_args[ERROR_CB_KEY] = self.obj.CredentialsError
        self.cred_args[DENIAL_CB_KEY] = self.obj.AuthorizationDenied

    def test_register(self):
        """The registration is correct."""
        self.obj.register(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)

    def test_login(self):
        """The login is correct."""
        self.obj.login(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)

    def test_login_email_password(self):
        """The login_email_password is correct."""
        self.args['email'] = EMAIL
        self.args['password'] = PASSWORD
        self.obj.login_email_password(APP_NAME, self.args)
        self.assert_recorder_called(self.factory, 'Credentials',
                                    APP_NAME, **self.cred_args)


class CredentialsManagementParamsTestCase(CredentialsManagementOpsTestCase):
    """Tests for the CredentialsManagement extra parameters handling."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementParamsTestCase, self).setUp()
        self.args['dummy'] = 'nothing useful'


class CredentialsManagementSignalsTestCase(CredentialsManagementTestCase):
    """Tests for the CredentialsManagement DBus signals."""

    def test_credentials_found(self):
        """The CredentialsFound signal."""
        self.obj.CredentialsFound(APP_NAME, TOKEN)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsFound.__name__, APP_NAME)
        self.assertTrue(self.memento.check_info(*msgs))

        msg = 'credentials must not be logged (found %r in log).'
        for val in TOKEN.itervalues():
            self.assertFalse(self.memento.check_info(val), msg % val)

    def test_credentials_not_found(self):
        """The CredentialsNotFound signal."""
        self.obj.CredentialsNotFound(APP_NAME)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsNotFound.__name__, APP_NAME)
        self.assertTrue(self.memento.check_info(*msgs))

    def test_credentials_cleared(self):
        """The CredentialsCleared signal."""
        self.obj.CredentialsCleared(APP_NAME)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsCleared.__name__, APP_NAME)
        self.assertTrue(self.memento.check_info(*msgs))

    def test_credentials_stored(self):
        """The CredentialsStored signal."""
        self.obj.CredentialsStored(APP_NAME)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsStored.__name__, APP_NAME)
        self.assertTrue(self.memento.check_info(*msgs))

    def test_credentials_error(self):
        """The CredentialsError signal."""
        error = {'error_message': 'failed!', 'detailed error': 'yadda yadda'}
        self.obj.CredentialsError(APP_NAME, error)
        msgs = (self.obj.__class__.__name__,
                self.obj.CredentialsError.__name__,
                APP_NAME, str(error))
        self.assertTrue(self.memento.check_error(*msgs))

    def test_authorization_denied(self):
        """The AuthorizationDenied signal."""
        self.obj.AuthorizationDenied(APP_NAME)
        msgs = (self.obj.__class__.__name__,
                self.obj.AuthorizationDenied.__name__, APP_NAME)
        self.assertTrue(self.memento.check_info(*msgs))


class UbuntuSSOServiceTestCase(BaseTestCase):
    """Test suite for the UbuntuSSOService class."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(UbuntuSSOServiceTestCase, self).setUp()
        self.proxy.sso_login = object()
        self.proxy.cred_manager = object()
        self.params = []
        self.patch(main, 'UbuntuSSOProxy',
                   lambda _: self.params.append(_) or self.proxy)
        self.obj = UbuntuSSOService()
        yield self.obj.start()

    def test_creation(self):
        """Creation paremeter is properly used."""
        self.assertEqual(self.params, [self.obj])

    def test_sso_login(self):
        """Attributes has the expected value."""
        self.assertIsInstance(self.obj.sso_login, SSOLogin)
        self.assertTrue(self.obj.sso_login.proxy is self.proxy.sso_login)

    def test_cred_manager(self):
        """Attributes has the expected value."""
        self.assertIsInstance(self.obj.cred_manager, CredentialsManagement)
        self.assertTrue(self.obj.cred_manager.proxy is self.proxy.cred_manager)
