# -*- coding: utf-8 -*-

# Authors: Natalia B. Bidart <natalia.bidart@canonical.com>
#          Diego Sarmentero <diego.sarmentero@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the Ui code."""

from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_sso import main


class FakedObject(object):
    """Fake an object, record every call."""

    next_result = None
    exposed_methods = []

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._called = {}
        for i in self.exposed_methods:
            setattr(self, i, self._record_call(i))

    def _record_call(self, func_name):
        """Store values when calling 'func_name'."""

        def inner(*args, **kwargs):
            """Fake 'func_name'."""
            self._called[func_name] = (args, kwargs)
            return self.next_result

        return inner


class FakedBackend(object):
    """A faked backend."""

    sso_login = FakedObject()


class FakePageUiStyle(object):
    """Fake the page."""

    def __init__(self):
        self.ui = self
        self.properties = {}
        super(FakePageUiStyle, self).__init__()

    def wizard(self):
        """Use itself as a fake wizard, too."""
        return self

    def text(self):
        """Return text."""
        return self.properties.get('text', '')

    # setText, setEnabled are inherited
    # pylint: disable=C0103
    def setText(self, text):
        """Save text."""
        self.properties['text'] = text

    def setEnabled(self, value):
        """Fake setEnabled."""
        self.properties['enabled'] = value

    def enabled(self):
        """Fake enabled."""
        return self.properties.get('enabled', False)

    def isEnabled(self):
        """Fake isEnabled."""
        return self.properties.get('enabled', False)

    def setProperty(self, key, val):
        """Fake setProperty to restyle some widget."""
        self.properties[key] = val

    def property(self, key):
        """Fake property from widget style."""
        return self.properties.get(key, False)

    def setDefault(self, val):
        """Fake button setDefault."""
        self.properties['default'] = val

    def style(self):
        """Fake style."""
        return self

    def unpolish(self, *args):
        """Fake unpolish."""
        self.properties['unpolish'] = len(args) > 0

    def polish(self, *args):
        """Fake polish."""
        self.properties['polish'] = len(args) > 0

    def setVisible(self, val):
        """Fake setVisible from Qt."""
        self.properties['visible'] = val

    def isVisible(self):
        """Fake isVisible from Qt."""
        return self.properties['visible']


class BaseTestCase(TestCase):
    """The base test case."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        self._called = False
        backend = FakedBackend()
        self.sso_login_backend = backend.sso_login
        self.patch(main, 'get_sso_client',
                   lambda *a: defer.succeed(backend))

    def _set_called(self, *args, **kwargs):
        """Store 'args' and 'kwargs' for test assertions."""
        self._called = (args, kwargs)
