# -*- coding: utf-8 -*-

# Authors: Manuel de la Pena <manuel@canonical.com>
#          Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the ui controllers."""

import collections

from twisted.internet import defer

from PyQt4.QtGui import QCheckBox, QLabel, QLineEdit, QWizard
from PyQt4.QtCore import QString, SIGNAL

# pylint: disable=F0401
try:
    from PIL import Image
except ImportError:
    import Image
# pylint: enable=F0401

from ubuntu_sso.qt.controllers import (
    BackendController,
    ChooseSignInController,
    CurrentUserController,
    EmailVerificationController,
    ErrorController,
    ForgottenPasswordController,
    FAKE_URL,
    ResetPasswordController,
    SetUpAccountController,
    SuccessController,
    UbuntuSSOWizardController,
)
from ubuntu_sso.qt.tests import BaseTestCase, FakePageUiStyle
from ubuntu_sso.utils.ui import (
    CAPTCHA_REQUIRED_ERROR,
    CAPTCHA_SOLUTION_ENTRY,
    EMAIL1_ENTRY,
    EMAIL2_ENTRY,
    EMAIL_INVALID,
    EMAIL_MISMATCH,
    ERROR,
    EMAIL_LABEL,
    EXISTING_ACCOUNT_CHOICE_BUTTON,
    FORGOTTEN_PASSWORD_BUTTON,
    JOIN_HEADER_LABEL,
    LOGIN_PASSWORD_LABEL,
    NAME_ENTRY,
    NAME_INVALID,
    PASSWORD1_ENTRY,
    PASSWORD2_ENTRY,
    PASSWORD_HELP,
    PASSWORD_TOO_WEAK,
    PASSWORD_MISMATCH,
    RESET_PASSWORD,
    REQUEST_PASSWORD_TOKEN_LABEL,
    SET_UP_ACCOUNT_CHOICE_BUTTON,
    SIGN_IN_BUTTON,
    SUCCESS,
    REQUEST_PASSWORD_TOKEN_WRONG_EMAIL,
    REQUEST_PASSWORD_TOKEN_TECH_ERROR,
    TRY_AGAIN_BUTTON,
    VERIFY_EMAIL_CONTENT,
    VERIFY_EMAIL_TITLE,
)

#ignore the comon mocker issues with lint
# pylint: disable=W0212,W0104,W0106,E1103


class FakeButton(FakePageUiStyle):

    """Fake Button."""

    def __init__(self):
        """Initialize object."""
        super(FakeButton, self).__init__()
        self.function = None
        self.clicked = self

    def connect(self, function):
        """Receive the function that should be connected to this button."""
        self.function = function


class FakeField(object):

    """Fake Field for text input."""

    def __init__(self, text):
        self.text = text

    # pylint: disable=C0103
    def toString(self):
        """Fake toString method for input fields."""
        return self.text
    # pylint: enable=C0103


class FakeWizard(object):

    """Fake Wizard."""

    current_user_page_id = 1
    setup_account_page_id = 2
    forgotten_password_page_id = 3
    reset_password_page_id = 4

    def __init__(self, view):
        """Initialize object."""
        self.properties = {}
        self.app_name = 'app_name'
        self.help_text = 'help'
        self.view = view
        self.next_id = 0
        # pylint: disable=C0103
        self.loginSuccess = self
        self.registrationSuccess = self
        # pylint: enable=C0103
        self.forgotten = self
        self.ui = self
        self.email_line_edit = QLineEdit()

    # pylint: disable=C0103
    def registerField(self, key, text):
        """Fake registerField from wizard."""
        self.properties['field_%s' % key] = FakeField(text)
    # pylint: enable=C0103

    def field(self, key):
        """Fake field from wizard"""
        return self.properties['field_%s' % key]

    def next(self):
        """Fake next method to move to to next page in the wizard."""
        self.view.properties['wizard_next'] = True

    def emit(self, name, email):
        """Fake emit."""
        self.properties['emit'] = (name, email)


class FakeGenericView(object):

    """Fake Generic page."""

    def __init__(self):
        """Initialize object."""
        self.properties = {}
        self.ui = self
        self.header = self
        self.properties['currentId'] = -1
        self.next = 0
        self.fake_wizard = FakeWizard(self)

    def set_title(self, title):
        """Fake set_title for the page."""
        self.properties['title'] = title

    def set_subtitle(self, subtitle):
        """Fake set_subtitle for the page"""
        self.properties['subtitle'] = subtitle

    # pylint: disable=C0103
    def setTitle(self, title):
        """Fake setTitle."""
        self.set_title(title)

    def setSubTitle(self, subtitle):
        """Fake setTitle."""
        self.set_subtitle(subtitle)
    # pylint: enable=C0103

    def wizard(self):
        """Fake wizard is returned."""
        return self.fake_wizard


class FakeChooseSignInView(FakeGenericView):

    """Fake ChooseSignIn page."""

    def __init__(self):
        """Initialize object."""
        super(FakeChooseSignInView, self).__init__()
        self.existing_account_button = FakeButton()
        self.setup_account_button = FakeButton()


class FakeCurrentUserView(FakeGenericView):

    """Fake Current User view."""

    def __init__(self):
        super(FakeCurrentUserView, self).__init__()
        self.email_label = QLabel()
        self.password_label = QLabel()
        self.forgot_password_label = QLabel()
        self.sign_in_button = FakeButton()
        self.email_edit = QLineEdit()
        self.password_edit = QLineEdit()
        self.on_logged_in_cb = None
        self.on_login_error_cb = None

    def fake_backend(self):
        """Fake get_backend."""
        return self

    def login(self, appname, email, password):
        """Fake backend login."""
        self.properties['backend_login'] = (appname, email, password)
        return self

    # pylint: disable=C0103
    def addErrback(self, function):
        """Fake defer addErrback."""
        self.properties['login-addErrback'] = function
    # pylint: enable=C0103


class FakeLineEdit(object):

    """A fake QLineEdit."""

    def __init__(self, *args):
        """Initialize."""
        self._text = u""

    def text(self):
        """Save text."""
        return self._text

    # setText is inherited
    # pylint: disable=C0103
    def setText(self, text):
        """Return saved text."""
        self._text = text


class FakePage(object):

    """A fake wizard page."""

    app_name = "APP"

    def wizard(self):
        """Fake wizard."""
        return self


class FakeCurrentUserPage(FakePageUiStyle):

    """Fake CurrentuserPage."""

    def __init__(self, *args):
        """Initialize."""
        super(FakeCurrentUserPage, self).__init__(*args)
        self.ui.email_edit = FakeLineEdit()
        self.ui.password_edit = FakeLineEdit()
        self.sign_in_button = self


class FakeSetupAccountPageView(FakeGenericView):

    """Fake SetupAccount view."""

    def __init__(self):
        super(FakeSetupAccountPageView, self).__init__()
        self.ui = self
        # Labels
        self.name_label = QLabel()
        self.email_label = QLabel()
        self.confirm_email_label = QLabel()
        self.password_label = QLabel()
        self.confirm_password_label = QLabel()
        self.password_info_label = QLabel()
        self.refresh_label = QLabel()
        self.name_assistance = QLabel()
        self.email_assistance = QLabel()
        self.confirm_email_assistance = QLabel()
        # Line Edits
        self.captcha_solution_edit = QLineEdit()
        self.name_edit = QLineEdit()
        self.email_edit = QLineEdit()
        self.confirm_email_edit = QLineEdit()
        self.password_edit = QLineEdit()
        self.confirm_password_edit = QLineEdit()
        self.captcha_solution_edit = QLineEdit()
        # Check Box
        self.terms_checkbox = QCheckBox()
        # Setup Button
        self.set_up_button = FakeButton()
        # Backend
        self.on_captcha_generated_cb = None
        self.on_captcha_generation_error_cb = None
        self.on_user_registered_cb = None
        self.on_user_registration_error_cb = None

    def fake_backend(self):
        """Fake get_backend."""
        return self

    def set_line_edit_validation_rule(self, edit, function):
        """Fake set_line_edit_validation_rule from view."""
        elements = self.properties.get('validation_rule', None)
        if elements is None:
            elements = []
        elements.append((edit, function))
        self.properties['validation_rule'] = elements

    # pylint: disable=C0103
    def registerField(self, key, edit):
        """Fake registerField from page."""
        self.properties[key] = edit
    # pylint: enable=C0103

    def set_error_message(self, label, message):
        """Fake set_error_message."""
        label.setText(message)


class FakeMessageBox(object):

    """A fake message box."""

    args = None
    critical_args = None

    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)

    def critical(self, *args, **kwargs):
        """Fake critical popup."""
        self.critical_args = (args, kwargs)


class FakeSetupAccountUi(object):
    """Fake ui variable for FakeView."""

    email_assistance = None


class FakeView(object):

    """A fake view."""

    app_name = "TestApp"
    ui = FakeSetupAccountUi()

    def wizard(self):
        """Use itself as a fake wizard, too."""
        return self

    def set_error_message(self, label, msg):
        """Simulate to add the message to the label as SetupAccount do."""


class FakeSetupAccountView(FakeView):

    """A Fake view."""

    captcha_file = None
    captcha_image = None
    captcha_refresh_executed = False
    captcha_refreshing_value = False

    def on_captcha_refresh_complete(self):
        """Fake the call to refresh finished."""
        self.captcha_refresh_executed = True

    def on_captcha_refreshing(self):
        """Fake the call to refreshing."""
        self.captcha_refreshing_value = True

    def fake_open(self, filename):
        """Fake open for Image."""
        return self

    # pylint: disable=W0622
    def save(self, io, format):
        """Fake save for Image."""
    # pylint: enable=W0622


class FakeControllerForCaptcha(object):

    """Fake controller to test refresh_captcha method."""

    def __init__(self):
        """Initialize FakeControllerForCaptcha."""
        self.callback_error = False

    def generate_captcha(self, *args):
        """Fake generate deferred for captcha."""
        return self

    # pylint: disable=C0103
    def addErrback(self, call):
        """Fake addErrback."""
        self.callback_error = call is not None
    # pylint: enable=C0103


class FakeEmailVerificationView(FakePageUiStyle):
    """Fake EmailVerification Page."""

    def __init__(self, *args):
        """Initialize."""
        super(FakeEmailVerificationView, self).__init__(*args)
        self.verification_code = ''
        self.next_button = self


class EmailVerificationView(FakeGenericView):

    """Fake Email Verification view."""

    def __init__(self):
        super(EmailVerificationView, self).__init__()
        self.ui = self
        # Line Edits
        self.verification_code_edit = QLineEdit()
        self.verification_code = ''
        # Setup Button
        self.next_button = FakeButton()
        # Backend
        self.on_email_validated_cb = None
        self.on_email_validation_error_cb = None

    def fake_backend(self):
        """Fake get_backend."""
        return self

    def validate_email(self, *args, **kwargs):
        """Fake validate_email method from backend"""
        self.properties['validate_email'] = (args, kwargs)


class ErrorPageView(FakeGenericView):

    """Fake Error Page view."""

    def __init__(self):
        super(ErrorPageView, self).__init__()
        self.ui = self
        self.error_message_label = QLabel()


class SuccessPageView(FakeGenericView):

    """Fake Success Page view."""

    def __init__(self):
        super(SuccessPageView, self).__init__()
        self.ui = self
        self.success_message_body = QLabel()


class UbuntuSSOView(object):

    """Fake view for UbuntuSSO."""

    success_page_id = 5
    error_page_id = 6

    def __init__(self):
        self.properties = {}
        self.ui = self
        self.app_name = 'app_name'
        self.success_message_label = QLabel()
        self.result = 0
        # pylint: disable=C0103
        self.loginSuccess = FakeButton()
        self.registrationSuccess = FakeButton()
        # pylint: enable=C0103
        self.page = FakeGenericView()

    def button(self, *args):
        """Fake button."""
        self.properties['button_method'] = args
        self.properties['button'] = FakeButton()
        return self.properties['button']

    def callback(self, *args, **kwargs):
        """Fake callback."""
        self.properties['callback'] = (args, kwargs)
        return self.result

    def close(self):
        """Fake close."""
        self.properties['close'] = True

    # pylint: disable=C0103
    def setButtonLayout(self, layout):
        """Fake setButtonLayout."""
        self.properties['buttons_layout'] = layout

    def currentPage(self):
        """Fake currentPage."""
        return self.page

    def setWizardStyle(self, style):
        """Fake setWizardStyle."""
        self.properties['wizard_style'] = style
    # pylint: enable=C0103

    def next(self):
        """Fake next method to move to to next page in the wizard."""
        self.properties['wizard_next'] = True


class FakeLineEditForForgottenPage(object):

    """Fake Line Edit"""

    def __init__(self):
        """Initilize object."""
        self.email_text = QString()

    def text(self):
        """Fake text for QLineEdit."""
        return self.email_text

    # setText is inherited
    # pylint: disable=C0103
    def setText(self, text):
        """Fake setText for QLineEdit."""
        self.email_text = QString(text)
    # pylint: enable=C0103


class FakeForgottenPasswordPage(FakePageUiStyle):
    """Fake the page."""

    def __init__(self):
        super(FakeForgottenPasswordPage, self).__init__()
        self.email_address_line_edit = self
        self.send_button = self
        self.email_widget = FakePageUiStyle()
        self.forgotted_password_intro_label = FakePageUiStyle()
        self.try_again_widget = FakePageUiStyle()
        self.email_line_edit = FakeLineEditForForgottenPage()

    def fake_backend(self):
        """Fake get_backend."""
        return self


class FakeForgottenPasswordPageView(FakeGenericView):
    """Fake the page."""

    def __init__(self):
        super(FakeForgottenPasswordPageView, self).__init__()
        self.email_address_line_edit = QLineEdit()
        self.send_button = FakeButton()
        self.try_again_button = FakeButton()
        self.email_widget = FakePageUiStyle()
        self.forgotted_password_intro_label = QLabel()
        self.email_address_label = QLabel()
        self.try_again_widget = FakePageUiStyle()
        self.email_line_edit = QLineEdit()
        # Backend
        self.on_password_reset_error_cb = None
        self.on_password_reset_token_sent_cb = None

    def fake_backend(self):
        """Fake get_backend."""
        return self

    # pylint: disable=C0103
    def registerField(self, key, item):
        """Fake registerField from wizard."""
        self.properties['field_%s' % key] = item
    # pylint: enable=C0103

    def field(self, key):
        """Fake field from wizard"""
        return self.properties['field_%s' % key]

    def set_line_edit_validation_rule(self, edit, function):
        """Fake set_line_edit_validation_rule from view."""
        elements = self.properties.get('validation_rule', None)
        if elements is None:
            elements = []
        elements.append((edit, function))
        self.properties['validation_rule'] = elements


class ResetPasswordPageView(FakeGenericView):
    """Fake the page."""

    def __init__(self):
        super(ResetPasswordPageView, self).__init__()
        self.reset_password_button = FakeButton()
        self.reset_code_line_edit = QLineEdit()
        self.password_line_edit = QLineEdit()
        self.confirm_password_line_edit = QLineEdit()
        # Backend
        self.on_password_changed_cb = None
        self.on_password_change_error_cb = None

    def fake_backend(self):
        """Fake get_backend."""
        return self

    # pylint: disable=C0103
    def registerField(self, key, item):
        """Fake registerField from wizard."""
        self.properties['field_%s' % key] = item
    # pylint: enable=C0103

    def field(self, key):
        """Fake field from wizard"""
        return self.properties['field_%s' % key]

    def set_line_edit_validation_rule(self, edit, function):
        """Fake set_line_edit_validation_rule from view."""
        elements = self.properties.get('validation_rule', None)
        if elements is None:
            elements = []
        elements.append((edit, function))
        self.properties['validation_rule'] = elements

    def set_new_password(self, *args, **kwargs):
        """Fake set_new_password from backend."""
        self.properties['backend_new_password'] = (args, kwargs)


class FakeResetPasswordPage(FakePageUiStyle):

    """Fake ResetPasswordPage."""

    def __init__(self, *args):
        """Initialize."""
        super(FakeResetPasswordPage, self).__init__()
        self.ui.reset_code_line_edit = FakeLineEdit()
        self.ui.password_line_edit = FakeLineEdit()
        self.ui.confirm_password_line_edit = FakeLineEdit()
        self.reset_password_button = self
        self._enabled = 9  # Intentionally wrong.


class FakeWizardForResetPassword(object):

    """Fake Wizard for ResetPasswordController."""

    def __init__(self):
        self.current_user = self
        self.ui = self
        self.email_edit = self
        self.email_line_edit = self
        self.forgotten = self
        self.overlay = self
        self.count_back = 0
        self.hide_value = False
        self.text_value = 'mail@mail.com'
        self.current_user_page_id = 4

    def wizard(self):
        """Fake wizard function for view."""
        return self

    def back(self):
        """Fake back for wizard."""
        self.count_back += 1

    def hide(self):
        """Fake hide for overlay."""
        self.hide_value = True

    def text(self):
        """Fake text for QLineEdit."""
        return self.text_value

    # pylint: disable=C0103
    def setText(self, text):
        """Fake setText for QLineEdit."""
        self.text_value = text

    def visitedPages(self):
        """Return an int list of fake visited pages."""
        return [1, 4, 6, 8]
    # pylint: enable=C0103


class BackendControllerTestCase(BaseTestCase):
    """The test case for the BackendController."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BackendControllerTestCase, self).setUp()
        self.backend = BackendController()

    @defer.inlineCallbacks
    def test_get_backend(self):
        """The backend is properly retrieved."""
        result = yield self.backend.get_backend()
        self.assertTrue(self.sso_login_backend is result)


class ChooseSignInControllerTestCase(BaseTestCase):
    """Test the choose sign in controller."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set tests."""
        yield super(ChooseSignInControllerTestCase, self).setUp()
        self.title = 'title'
        self.subtitle = 'subtitle'
        self.controller = ChooseSignInController(self.title, self.subtitle)
        self.view = FakeChooseSignInView()
        self.controller.view = self.view

    def test_setup_ui(self):
        """Test the set up of the ui."""
        self.controller.setupUi(self.view)
        self.assertEqual(self.view.properties['title'], self.title)
        self.assertEqual(self.view.properties['subtitle'], self.subtitle)
        self.assertEqual(self.view.existing_account_button.text(),
            EXISTING_ACCOUNT_CHOICE_BUTTON)
        self.assertEqual(self.view.setup_account_button.text(),
            SET_UP_ACCOUNT_CHOICE_BUTTON)
        self.assertIsInstance(self.view.existing_account_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.setup_account_button.function,
            collections.Callable)

    def test_set_up_translated_strings(self):
        """Ensure that the translations are used."""
        self.controller._set_up_translated_strings()
        self.assertEqual(self.view.existing_account_button.text(),
            EXISTING_ACCOUNT_CHOICE_BUTTON)
        self.assertEqual(self.view.setup_account_button.text(),
            SET_UP_ACCOUNT_CHOICE_BUTTON)

    def test_connect_buttons(self):
        """Ensure that all the buttons are correcly connected."""
        self.controller._connect_buttons()
        self.assertIsInstance(self.view.existing_account_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.setup_account_button.function,
            collections.Callable)

    def test_set_next_existing(self):
        """Test the execution of the callback."""
        self.controller._set_next_existing()
        self.assertEqual(self.view.next,
            self.view.fake_wizard.current_user_page_id)

    def test_set_next_new(self):
        """Test the execution of the callback."""
        self.controller._set_next_new()
        self.assertEqual(self.view.next,
            self.view.fake_wizard.setup_account_page_id)


class CurrentUserControllerTestCase(BaseTestCase):
    """Test the current user controller."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup tests."""
        yield super(CurrentUserControllerTestCase, self).setUp()
        self.controller = CurrentUserController(title='the title',
            subtitle='the subtitle')
        self.view = FakeCurrentUserView()
        self.controller.view = self.view
        self.patch(self.controller, "get_backend", self.view.fake_backend)
        self.controller.setupUi(self.view)

    def test_translated_strings(self):
        """test that the ui is correctly set up."""
        self.controller._set_translated_strings()
        self.assertEqual(self.view.email_label.text(), EMAIL_LABEL)
        self.assertEqual(self.view.password_label.text(), LOGIN_PASSWORD_LABEL)
        self.assertEqual(self.view.forgot_password_label.text(),
            FAKE_URL % FORGOTTEN_PASSWORD_BUTTON)
        self.assertEqual(self.view.sign_in_button.text(), SIGN_IN_BUTTON)

    def test_connect_ui(self):
        """test that the ui is correctly set up."""
        self.controller._connect_ui()
        self.assertIsInstance(self.view.sign_in_button.function,
            collections.Callable)
        # Check if receivers is 2 because _connect_ui was called on setUp
        self.assertEqual(self.view.forgot_password_label.receivers(
            SIGNAL('linkActivated(QString)')), 2)
        self.assertEqual(self.view.email_edit.receivers(
            SIGNAL('textChanged(QString)')), 2)
        self.assertEqual(self.view.password_edit.receivers(
            SIGNAL('textChanged(QString)')), 2)
        self.assertFalse(self.view.on_login_error_cb is None)
        self.assertFalse(self.view.on_logged_in_cb is None)

    def test_title_subtitle(self):
        """Ensure we are storing the title and subtitle correctly."""
        self.assertEqual(self.view.properties['title'], 'the title')
        self.assertEqual(self.view.properties['subtitle'], 'the subtitle')

    def test_on_logged_in(self):
        """Test on_logged_in method."""
        self.view.email_edit.setText('email@email.com')
        self.controller.on_logged_in('app-name', {})
        self.assertEqual(self.view.fake_wizard.properties['emit'],
            ('app-name', 'email@email.com'))

    def test_login(self):
        """Test login method."""
        email = 'email@email.com'
        password = 'password'
        self.view.email_edit.setText(email)
        self.view.password_edit.setText(password)
        self.controller.login()
        self.assertEqual(self.view.properties['backend_login'],
            (self.view.fake_wizard.app_name, email, password))
        self.assertIsInstance(self.view.properties['login-addErrback'],
            collections.Callable)

    def test_on_forgotten_password(self):
        """Test on_forgotten_password flow."""
        email = 'email@email.com'
        self.view.email_edit.setText(email)
        self.controller.on_forgotten_password()
        self.assertEqual(self.view.next,
            self.view.fake_wizard.forgotten_password_page_id)
        self.assertTrue(self.view.properties['wizard_next'])

    def test_setup_ui(self):
        """Test the set up of the ui."""
        self.controller._title = 'title_setup'
        self.controller._subtitle = 'subtitle_setup'
        self.controller.setupUi(self.view)
        self.assertEqual(self.view.properties['title'], 'title_setup')
        self.assertEqual(self.view.properties['subtitle'], 'subtitle_setup')
        self.assertIsInstance(self.view.sign_in_button.function,
            collections.Callable)
        # Check if receivers is 2 because _connect_ui was called on setUp
        self.assertEqual(self.view.forgot_password_label.receivers(
            SIGNAL('linkActivated(QString)')), 2)
        self.assertEqual(self.view.email_edit.receivers(
            SIGNAL('textChanged(QString)')), 2)
        self.assertEqual(self.view.password_edit.receivers(
            SIGNAL('textChanged(QString)')), 2)
        self.assertFalse(self.view.on_login_error_cb is None)
        self.assertFalse(self.view.on_logged_in_cb is None)
        self.assertEqual(self.view.email_label.text(), EMAIL_LABEL)
        self.assertEqual(self.view.password_label.text(), LOGIN_PASSWORD_LABEL)
        self.assertEqual(self.view.forgot_password_label.text(),
            FAKE_URL % FORGOTTEN_PASSWORD_BUTTON)
        self.assertEqual(self.view.sign_in_button.text(), SIGN_IN_BUTTON)


class CurrentUserControllerErrorTestCase(BaseTestCase):

    """Tests for CurrentUserController's error handler."""

    on_error_method_name = "on_login_error"
    controller_class = CurrentUserController

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(CurrentUserControllerErrorTestCase, self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = self.controller_class(
            message_box=self.message_box)
        self.controller.view = FakePage()
        self.on_error_method = getattr(
            self.controller, self.on_error_method_name)

    def test_error_message_key(self):
        """Test that on_login_error reacts to errors with "error_message"."""
        self.on_error_method({"error_message": "WORRY!"})
        self.assertEqual(self.message_box.critical_args, (('WORRY!',
            self.controller.view), {}))

    def test_message_key(self):
        """Test that on_login_error reacts to errors with "message"."""
        self.on_error_method({"message": "WORRY!"})
        self.assertEqual(self.message_box.critical_args, (('WORRY!',
            self.controller.view), {}))

    def test_broken_error(self):
        """Test that on_login_error reacts to broken errors."""
        self.on_error_method({"boo!": "WORRY!"})
        result = '\n'.join(
                [('%s: %s' % (k, v)) for k, v in \
                {"boo!": "WORRY!"}.iteritems()])
        self.assertEqual(self.message_box.critical_args,
            ((result, self.controller.view), {}))

    def test_all_and_message(self):
        """Test that on_login_error reacts to broken errors."""
        self.on_error_method(
            {"message": "WORRY!", "__all__": "MORE!"})
        self.assertEqual(self.message_box.critical_args,
            (('MORE!\nWORRY!', self.controller.view), {}))

    def test_all_and_error_message(self):
        """Test that on_login_error reacts to broken errors."""
        self.on_error_method(
            {"error_message": "WORRY!", "__all__": "MORE!"})
        self.assertEqual(self.message_box.critical_args,
            (('MORE!\nWORRY!', self.controller.view), {}))

    def test_only_all(self):
        """Test that on_login_error reacts to broken errors."""
        self.on_error_method(
            {"__all__": "MORE!"})
        self.assertEqual(self.message_box.critical_args,
            (('MORE!', self.controller.view), {}))


class EmailVerificationControllerErrorTestCase(
    CurrentUserControllerErrorTestCase):

    """Tests for EmailVerificationController's error handler."""

    on_error_method_name = "on_email_validation_error"
    controller_class = EmailVerificationController

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(EmailVerificationControllerErrorTestCase, self).setUp()
        # This error handler takes one extra argument.
        self.on_error_method = lambda error: getattr(
            self.controller, self.on_error_method_name)('APP', error)


class SetUpAccountControllerErrorTestCase(
    EmailVerificationControllerErrorTestCase):

    """Tests for SetUpAccountController's error handler."""

    on_error_method_name = "on_user_registration_error"
    controller_class = SetUpAccountController

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(SetUpAccountControllerErrorTestCase, self).setUp()
        self.patch(self.controller, "_refresh_captcha", lambda *args: None)


class ResetPasswordControllerErrorTestCase(
    EmailVerificationControllerErrorTestCase):

    """Tests for ResetPasswordController's error handler."""

    on_error_method_name = "on_password_change_error"
    controller_class = ResetPasswordController


class CurrentUserControllerValidationTest(BaseTestCase):

    """Tests for CurrentUserController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(CurrentUserControllerValidationTest, self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = CurrentUserController(
            message_box=self.message_box)
        self.controller.view = FakeCurrentUserPage()

    def test_valid(self):
        """Enable the button with a valid email/password."""
        self.controller.view.email_edit.setText("a@b")
        self.controller.view.password_edit.setText("pass")
        self.controller._validate()
        self.assertTrue(self.controller.view.sign_in_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.sign_in_button.enabled())

    def test_invalid_email(self):
        """The submit button should be disabled with an invalid email."""
        self.controller.view.email_edit.setText("ab")
        self.controller.view.password_edit.setText("pass")
        self.controller._validate()
        self.assertFalse(self.controller.view.sign_in_button.enabled())
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.sign_in_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_invalid_password(self):
        """The submit button should be disabled with an invalid password."""
        self.controller.view.email_edit.setText("a@b")
        self.controller.view.password_edit.setText("")
        self.controller._validate()
        self.assertFalse(self.controller.view.sign_in_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.sign_in_button.enabled())

    def test_invalid_both(self):
        """The submit button should be disabled with invalid data."""
        self.controller.view.email_edit.setText("ab")
        self.controller.view.password_edit.setText("")
        self.controller._validate()
        self.assertFalse(self.controller.view.sign_in_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.sign_in_button.enabled())


class SetUpAccountControllerTestCase(BaseTestCase):
    """test the controller used to setup a new account."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(SetUpAccountControllerTestCase, self).setUp()
        self.view = FakeSetupAccountPageView()
        self.message_box = FakeMessageBox()
        self.controller = SetUpAccountController(message_box=self.message_box)
        self.patch(self.controller, "get_backend", self.view.fake_backend)
        self.controller.view = self.view
        self.controller.backend = self.view

    def test_set_translated_strings(self):
        """Ensure all the strings are set."""
        self.controller._set_translated_strings()
        self.assertEqual(self.view.name_label.text(), NAME_ENTRY)
        self.assertEqual(self.view.email_label.text(), EMAIL1_ENTRY)
        self.assertEqual(self.view.confirm_email_label.text(), EMAIL2_ENTRY)
        self.assertEqual(self.view.password_label.text(), PASSWORD1_ENTRY)
        self.assertEqual(self.view.confirm_password_label.text(),
            PASSWORD2_ENTRY)
        self.assertEqual(self.view.password_info_label.text(), PASSWORD_HELP)
        self.assertEqual(self.view.captcha_solution_edit.placeholderText(),
            CAPTCHA_SOLUTION_ENTRY)

    def test_set_titles(self):
        """Test how the different titles are set."""
        self.controller._set_titles()
        self.assertEqual(self.view.properties['title'],
            JOIN_HEADER_LABEL % {'app_name': self.view.fake_wizard.app_name})
        self.assertEqual(self.view.properties['subtitle'],
            self.view.fake_wizard.help_text)

    def test_connect_ui_elements(self):
        """Test that the ui elements are correctly connect."""
        self.controller._connect_ui_elements()
        self.assertEqual(self.view.name_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.email_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.confirm_email_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.password_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.confirm_password_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.captcha_solution_edit.receivers(
            SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.view.refresh_label.receivers(
            SIGNAL('linkActivated(QString)')), 1)
        self.assertEqual(self.view.terms_checkbox.receivers(
            SIGNAL('stateChanged(int)')), 1)
        # set the callbacks for the captcha generation
        self.assertIsInstance(self.view.on_captcha_generated_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_captcha_generation_error_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_user_registration_error_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_user_registered_cb,
            collections.Callable)

    def test_set_line_edits_validations(self):
        """Test _set_line_validations from controller."""
        self.controller._set_line_edits_validations()
        elements = self.view.properties['validation_rule']
        self.assertFalse(elements is None)
        for e in elements:
            self.assertTrue(type(e[0]) is QLineEdit)
            self.assertIsInstance(e[1], collections.Callable)

    def all_valid(self):
        """Set all the widgets to a valid state."""
        self.view.name_edit.setText('Name')
        self.view.email_edit.setText('email@email.com')
        self.view.confirm_email_edit.setText('email@email.com')
        self.view.password_edit.setText('T3st3rqw')
        self.view.confirm_password_edit.setText('T3st3rqw')
        self.view.captcha_solution_edit.setText('captcha solution')
        self.view.terms_checkbox.setChecked(True)

    def all_invalid(self):
        """Set all the widgets to an invalid state."""
        self.view.name_edit.setText('')
        self.view.email_edit.setText('')
        self.view.confirm_email_edit.setText('email')
        self.view.password_edit.setText('')
        self.view.confirm_password_edit.setText('t')
        self.view.captcha_solution_edit.setText('')
        self.view.terms_checkbox.setChecked(False)

    def test_enable_setup_button_only_checkbox_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.terms_checkbox.setChecked(True)
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_name_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.name_edit.setText('Name')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_email_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.email_edit.setText('email@email.com')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_confirm_email_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.confirm_email_edit.setText('email@email.com')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_email_and_confirm_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.email_edit.setText('email@email.com')
        self.view.confirm_email_edit.setText('email@email.com')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_password_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.password_edit.setText('T3st3rqw')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_password_confirm_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.confirm_password_edit.setText('T3st3rqw')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_password_and_confirm_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.password_edit.setText('T3st3rqw')
        self.view.confirm_password_edit.setText('T3st3rqw')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_captcha_ok(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.view.captcha_solution_edit.setText('captcha solution')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_only_all_ok(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.controller._enable_setup_button()

        self.assertTrue(self.view.set_up_button.enabled())
        self.assertFalse(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_all_wrong(self):
        """Test enable button only name valid."""
        self.all_invalid()
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_name_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.name_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_email_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.email_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_confirm_email_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.confirm_email_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_password_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.password_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_confirm_password_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.confirm_password_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_captcha_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.captcha_solution_edit.setText('')
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_enable_setup_button_checkbox_wrong(self):
        """Test enable button only name valid."""
        self.all_valid()
        self.view.terms_checkbox.setChecked(False)
        self.controller._enable_setup_button()

        self.assertFalse(self.view.set_up_button.enabled())
        self.assertTrue(self.view.set_up_button.property('DisabledState'))
        self.assertTrue(self.view.set_up_button.properties['polish'])
        self.assertTrue(self.view.set_up_button.properties['unpolish'])

    def test_register_fields(self):
        """Test _register_fields."""
        self.controller._register_fields()
        self.assertTrue('email_address' in self.view.properties)
        self.assertTrue('password' in self.view.properties)
        self.assertTrue(type(self.view.properties['email_address']) is \
            QLineEdit)
        self.assertTrue(type(self.view.properties['password']) is \
            QLineEdit)

    def test_validate_form_all_ok(self):
        """Test the result of validate_form."""
        self.all_valid()
        self.assertTrue(self.controller.validate_form())
        self.assertNotEqual(self.view.name_assistance.text(), NAME_INVALID)
        self.assertNotEqual(self.view.email_assistance.text(), EMAIL_INVALID)
        self.assertNotEqual(self.view.confirm_email_assistance.text(),
            EMAIL_MISMATCH)
        messages = '\n'.join([PASSWORD_TOO_WEAK, PASSWORD_MISMATCH,
            CAPTCHA_REQUIRED_ERROR])
        self.assertNotEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_all_wrong(self):
        """Test the result of validate_form."""
        self.all_invalid()
        self.assertFalse(self.controller.validate_form())
        self.assertEqual(self.view.name_assistance.text(), NAME_INVALID)
        self.assertEqual(self.view.email_assistance.text(), EMAIL_INVALID)
        self.assertEqual(self.view.confirm_email_assistance.text(),
            EMAIL_MISMATCH)
        messages = '\n'.join([PASSWORD_TOO_WEAK, PASSWORD_MISMATCH,
            CAPTCHA_REQUIRED_ERROR])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_name_ok(self):
        """Test the result of validate_form."""
        self.all_invalid()
        self.view.name_edit.setText('Name')
        self.assertFalse(self.controller.validate_form())
        self.assertEqual(self.view.email_assistance.text(), EMAIL_INVALID)
        self.assertEqual(self.view.confirm_email_assistance.text(),
            EMAIL_MISMATCH)
        messages = '\n'.join([PASSWORD_TOO_WEAK, PASSWORD_MISMATCH,
            CAPTCHA_REQUIRED_ERROR])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_email_ok(self):
        """Test the result of validate_form."""
        self.all_invalid()
        self.view.email_edit.setText('email@email.com')
        self.assertFalse(self.controller.validate_form())
        self.assertEqual(self.view.name_assistance.text(), NAME_INVALID)
        self.assertEqual(self.view.confirm_email_assistance.text(),
            EMAIL_MISMATCH)
        messages = '\n'.join([PASSWORD_TOO_WEAK, PASSWORD_MISMATCH,
            CAPTCHA_REQUIRED_ERROR])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_confirm_email_ok(self):
        """Test the result of validate_form."""
        self.all_invalid()
        self.view.email_edit.setText('email@email.com')
        self.view.confirm_email_edit.setText('email@email.com')
        self.assertFalse(self.controller.validate_form())
        self.assertEqual(self.view.name_assistance.text(), NAME_INVALID)
        messages = '\n'.join([PASSWORD_TOO_WEAK, PASSWORD_MISMATCH,
            CAPTCHA_REQUIRED_ERROR])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_password_weak(self):
        """Test the result of validate_form."""
        self.all_valid()
        self.view.password_edit.setText('Test')
        self.view.confirm_password_edit.setText('Test')
        self.assertFalse(self.controller.validate_form())
        messages = '\n'.join([PASSWORD_TOO_WEAK])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_password_mismatch(self):
        """Test the result of validate_form."""
        self.all_valid()
        self.view.password_edit.setText('T3st3rqw')
        self.view.confirm_password_edit.setText('test')
        self.assertFalse(self.controller.validate_form())
        messages = '\n'.join([PASSWORD_MISMATCH])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_validate_form_captcha_required(self):
        """Test the result of validate_form."""
        self.all_valid()
        self.view.captcha_solution_edit.setText('')
        self.assertFalse(self.controller.validate_form())
        messages = '\n'.join([CAPTCHA_REQUIRED_ERROR])
        self.assertEqual(self.message_box.critical_args,
            ((messages, self.view), {}))

    def test_is_correct_email(self):
        """Test if the email is correct."""
        self.assertTrue(self.controller.is_correct_email('email@'))
        self.assertFalse(self.controller.is_correct_email('email'))

    def test_is_correct_email_confirmation(self):
        """Test if the email confirmation is correct."""
        self.view.email_edit.setText('email@email.com')
        self.controller.is_correct_email_confirmation('email.@email.com')

    def test_is_correct_password_confirmation(self):
        """Test is_correct_password_confirmation method."""
        self.view.password_edit.setText('T3st3rqw')
        self.controller.is_correct_password_confirmation('T3st3rqw')


class SetupAccountControllerCaptchaTest(BaseTestCase):
    """Tests for SetupAccountController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(SetupAccountControllerCaptchaTest, self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = SetUpAccountController(message_box=self.message_box)
        self.patch(self.controller, 'view', FakeSetupAccountView())
        self.fake_backend = FakeControllerForCaptcha()
        self.patch(self.controller, 'backend', self.fake_backend)

    def test_refresh_captcha(self):
        """Test the Refresh Captcha function."""
        self.assertFalse(self.controller.view.captcha_refreshing_value)
        self.controller._refresh_captcha()
        self.assertTrue(self.controller.view.captcha_refreshing_value)
        self.assertTrue(self.fake_backend.callback_error)


class SetupAccountControllerValidationTest(BaseTestCase):
    """Tests for SetupAccountController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(SetupAccountControllerValidationTest, self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = SetUpAccountController(message_box=self.message_box)
        self.patch(self.controller, '_refresh_captcha', self._set_called)
        self.patch(self.controller, 'view', FakeSetupAccountView())

    def test_on_user_registration_refresh_captcha(self):
        """If there is a user reg. error, captcha should refresh."""
        self.controller.on_user_registration_error('TestApp', {})
        self.assertEqual(self._called, ((), {}))

    def test_on_user_registration_all_only(self):
        """Pass only a __all__ error key."""
        self.controller.on_user_registration_error('TestApp',
            {'__all__': "Error in All"})
        self.assertEqual(self.message_box.critical_args, ((
            "Error in All", self.controller.view), {}))

    def test_on_user_registration_all_fields(self):
        """Pass all known error keys, plus unknown one."""
        self.controller.on_user_registration_error('TestApp',
            {'__all__': "Error in All",
             'email': "Error in email",
             'pasword': "Error in password",
             'unknownfield': "Error in unknown",
            })
        self.assertEqual(self.message_box.critical_args, ((
            "Error in All", self.controller.view), {}))

    def test_registration_errors_without_message_or_all(self):
        """Pass only a email error key."""
        errdict = {'errtype': "RegistrationError",
                   'email': "Error in email"}
        self.controller.on_user_registration_error('TestApp', errdict)

        expected = (('', self.controller.view), {})
        self.assertEqual(self.message_box.critical_args, expected)

    def test_on_captcha_generated(self):
        """Test if the method that shows the overlay is executed."""
        self.patch(Image, "open", self.controller.view.fake_open)
        self.assertFalse(self.controller.view.captcha_refresh_executed)
        self.controller.on_captcha_generated('app_name', 'captcha_id')
        self.assertTrue(self.controller.view.captcha_refresh_executed)

    def test_on_captcha_generation_error(self):
        """Test if the method that hides the overlay is executed."""
        self.assertFalse(self.controller.view.captcha_refresh_executed)
        self.controller.on_captcha_generation_error({})
        self.assertTrue(self.controller.view.captcha_refresh_executed)


class EmailVerificationControllerTestCase(BaseTestCase):
    """Test the controller."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set tests."""
        yield super(EmailVerificationControllerTestCase, self).setUp()
        self.view = EmailVerificationView()
        self.backend = self.view
        self.controller = EmailVerificationController(
            message_box=FakeMessageBox())
        self.controller.view = self.view
        self.controller.backend = self.backend
        self.patch(self.controller, "get_backend", self.view.fake_backend)
        self.email = 'email@email.com'
        self.password = 'T3st3rqw'
        self.view.fake_wizard.registerField('email_address', self.email)
        self.view.fake_wizard.registerField('password', self.password)

    def test_connect_ui_elements(self):
        """Set the ui connections."""
        self.controller._connect_ui_elements()
        self.assertEqual(self.view.verification_code_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)
        self.assertIsInstance(self.view.next_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.on_email_validated_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_email_validation_error_cb,
            collections.Callable)

    def test_set_titles(self):
        """Test that the titles are set."""
        self.controller.set_titles()
        self.assertEqual(self.view.properties['title'], VERIFY_EMAIL_TITLE)
        self.assertEqual(self.view.properties['subtitle'],
            VERIFY_EMAIL_CONTENT % {
            "app_name": self.view.fake_wizard.app_name,
            "email": self.email,
            })

    def test_validate_email(self):
        """Test the callback."""
        code = 'qwe123'
        self.view.verification_code_edit.setText(code)
        self.controller.validate_email()
        self.assertEqual(self.view.properties['validate_email'],
            ((self.view.fake_wizard.app_name, self.email, self.password,
            code), {}))

    def test_validate_form(self):
        """Test validate_form."""
        self.view.verification_code = 'qwe123'
        self.controller.validate_form()
        self.assertTrue(self.view.next_button.enabled())
        self.assertFalse(self.view.next_button.property('DisabledState'))
        self.assertTrue(self.view.next_button.properties['polish'])
        self.assertTrue(self.view.next_button.properties['unpolish'])

        self.view.verification_code = ''
        self.controller.validate_form()
        self.assertFalse(self.view.next_button.enabled())
        self.assertTrue(self.view.next_button.property('DisabledState'))
        self.assertTrue(self.view.next_button.properties['polish'])
        self.assertTrue(self.view.next_button.properties['unpolish'])

    def test_page_initialized(self):
        """test pageInitialized to check the initial state of the page."""
        self.controller.pageInitialized()
        self.assertTrue(self.view.next_button.properties['default'])
        self.assertFalse(self.view.next_button.enabled())
        self.assertTrue(self.view.next_button.property('DisabledState'))
        self.assertTrue(self.view.next_button.properties['polish'])
        self.assertTrue(self.view.next_button.properties['unpolish'])

    def test_on_email_validation_error(self):
        """Test on_email_validation_error."""
        error = dict(error='email error')
        self.controller.on_email_validation_error('app', error)
        result = '\n'.join(
                [('%s: %s' % (k, v)) for k, v in error.iteritems()])
        self.assertEqual(self.controller.message_box.critical_args,
            ((result, self.view), {}))

    def test_on_email_validated(self):
        """Test on_email_validated."""
        self.controller.on_email_validated('app_name')
        self.assertEqual(self.view.fake_wizard.properties['emit'],
            ('app_name', self.email))


class EmailVerificationControllerValidationTestCase(BaseTestCase):
    """Tests for EmailVerificationController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(EmailVerificationControllerValidationTestCase,
            self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = EmailVerificationController(
            message_box=self.message_box)
        self.patch(self.controller, 'view', FakeEmailVerificationView())

    def test_on_email_validation_error(self):
        """Test that on_email_validation_error callback works as expected."""
        # Error type is removed from the final message
        error = dict(errtype='BadTokenError')
        app_name = 'app_name'
        self.controller.on_email_validation_error(app_name, error)
        self.assertEqual(self.message_box.critical_args,
            (('', self.controller.view), {}))

    def test_validate_form_wrong(self):
        """Check the state of the next button."""
        self.controller.view.verification_code = ''
        self.controller.validate_form()
        self.assertFalse(self.controller.view.next_button.isEnabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.next_button.enabled())

    def test_validate_form_ok(self):
        """Check the state of the next button."""
        self.controller.view.verification_code = 'as322fdw'
        self.controller.validate_form()
        self.assertTrue(self.controller.view.next_button.isEnabled)
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.next_button.enabled())


class ErrorControllerTestCase(BaseTestCase):
    """Test the success page controller."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(ErrorControllerTestCase, self).setUp()
        self.view = ErrorPageView()
        self.backend = self.view
        self.controller = ErrorController()
        self.controller.view = self.view
        self.controller.backend = self.backend

    def test_set_ui(self):
        """Test the process that sets the ui."""
        self.controller._title = ERROR
        self.controller._subtitle = ERROR
        self.controller.setupUi(self.view)
        self.assertEqual(self.view.error_message_label.text(), ERROR)
        self.assertEqual(self.view.properties['title'], ERROR)
        self.assertEqual(self.view.properties['subtitle'], ERROR)
        self.assertEqual(self.view.next, -1)


class SuccessControllerTestCase(BaseTestCase):
    """Test the success page controller."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SuccessControllerTestCase, self).setUp()
        self.view = SuccessPageView()
        self.backend = self.view
        self.controller = SuccessController()
        self.controller.view = self.view
        self.controller.backend = self.backend

    def test_set_ui(self):
        """Test the process that sets the ui."""
        self.controller._title = SUCCESS
        self.controller._subtitle = SUCCESS
        self.controller.setupUi(self.view)
        self.assertEqual(self.view.properties['title'], SUCCESS)
        self.assertEqual(self.view.properties['subtitle'], SUCCESS)
        self.assertEqual(self.view.next, -1)


class UbuntuSSOWizardControllerTestCase(BaseTestCase):
    """Test the wizard controller."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set tests."""
        yield super(UbuntuSSOWizardControllerTestCase, self).setUp()
        self.view = UbuntuSSOView()
        self.backend = self.view
        self.callback = self.view.callback
        self.controller = UbuntuSSOWizardController()
        self.controller.view = self.view
        self.controller.backend = self.backend

    def test_on_user_cancelation(self):
        """Test that the callback is indeed called."""
        self.controller.user_cancellation_callback = self.callback
        self.controller.on_user_cancelation()
        self.assertTrue(self.view.properties.get('close', False))
        self.assertEqual(self.view.properties['callback'],
            ((self.view.app_name, ), {}))

    def test_on_login_success(self):
        """Test that the callback is indeed called."""
        app_name = 'app'
        email = 'email'
        self.controller.login_success_callback = self.callback
        self.controller.login_success_callback(app_name, email)
        self.assertEqual(self.view.properties['callback'],
            ((app_name, email), {}))

    def test_on_registration_success(self):
        """Test that the callback is indeed called."""
        app_name = 'app'
        email = 'email'
        self.controller.registration_success_callback = self.callback
        self.controller.registration_success_callback(app_name, email)
        self.assertEqual(self.view.properties['callback'],
            ((app_name, email), {}))

    def test_show_success_message(self):
        """Test that the correct page will be shown."""
        self.controller.show_success_message()
        # the buttons layout we expect to have
        layout = []
        layout.append(QWizard.Stretch)
        layout.append(QWizard.FinishButton)

        self.assertEqual(self.view.page.next, self.view.success_page_id)
        self.assertTrue(self.view.properties['wizard_next'])
        self.assertEqual(self.view.properties['buttons_layout'], layout)

    def test_show_error_message(self):
        """Test that the correct page will be shown."""
        self.controller.show_error_message()
        # the buttons layout we expect to have
        layout = []
        layout.append(QWizard.Stretch)
        layout.append(QWizard.FinishButton)

        self.assertEqual(self.view.page.next, self.view.error_page_id)
        self.assertTrue(self.view.properties['wizard_next'])
        self.assertEqual(self.view.properties['buttons_layout'], layout)

    def test_setup_ui(self):
        """Test that the ui is connect."""
        self.controller.setupUi(self.view)
        self.assertEqual(self.view.properties['wizard_style'],
            QWizard.ModernStyle)
        self.assertIsInstance(self.view.properties['button'].function,
            collections.Callable)
        self.assertIsInstance(self.view.loginSuccess.function,
            collections.Callable)
        self.assertIsInstance(self.view.registrationSuccess.function,
            collections.Callable)


class ForgottenPasswordControllerValidationTest(BaseTestCase):

    """Tests for ForgottenPasswordController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(ForgottenPasswordControllerValidationTest, self).setUp()
        self.message_box = FakeMessageBox()
        self.controller = ForgottenPasswordController(
            message_box=self.message_box)
        self.view = FakeForgottenPasswordPage()
        self.controller.view = self.view
        self.patch(self.controller, "get_backend", self.view.fake_backend)

    def test_page_initialized(self):
        """Test the initial state of the page when it is loaded."""
        self.controller.pageInitialized()
        self.assertFalse(self.controller.view.send_button.enabled())
        self.assertTrue(self.controller.view.send_button.properties['default'])
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.send_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_page_initialized_with_text(self):
        """Test the initial state of the page when it is loaded."""
        self.controller.view.email_line_edit.setText('mail@mail.com')
        self.controller.pageInitialized()
        self.assertTrue(self.controller.view.send_button.enabled())
        self.assertTrue(
            self.controller.view.send_button.properties['default'])
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.send_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_valid(self):
        """The submit button should be enabled with a valid email."""
        self.controller.view.email_address_line_edit.setText("a@b")
        self.assertNotEqual(unicode(
            self.controller.view.email_address_line_edit.text()), u"")
        self.controller._validate()
        self.assertTrue(self.controller.view.send_button.enabled())
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.send_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_invalid(self):
        """The submit button should be disabled with an invalid email."""
        self.controller.view.email_address_line_edit.setText("ab")
        self.assertNotEqual(
            unicode(self.controller.view.email_address_line_edit.text()), u"")
        self.controller._validate()
        self.assertFalse(self.controller.view.send_button.enabled())
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.send_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_empty(self):
        """The submit button should be disabled without email."""
        self.assertEqual(
            unicode(self.controller.view.email_address_line_edit.text()), u"")
        self.assertFalse(self.controller.view.send_button.enabled())

    def test_on_password_reset_error_token_error(self):
        """Test that the on_password_reset_error callback works as expected."""
        error = dict(errtype='ResetPasswordTokenError')
        app_name = 'app_name'
        self.controller.on_password_reset_error(app_name, error)
        msg = REQUEST_PASSWORD_TOKEN_WRONG_EMAIL
        self.assertEqual(self.controller.message_box.critical_args,
            ((msg, self.controller.view), {}))

    def test_on_password_reset_error_general_error(self):
        """Test that the on_password_reset_error callback works as expected."""
        error = dict(errtype='RandomError')
        app_name = 'app_name'
        msg = REQUEST_PASSWORD_TOKEN_TECH_ERROR
        self.controller.on_password_reset_error(app_name, error)
        self.assertFalse(self.controller.view.email_widget.isVisible())
        self.assertFalse(
            self.controller.view.forgotted_password_intro_label.isVisible())
        self.assertTrue(self.controller.view.try_again_widget.isVisible())
        self.assertEqual(self.controller.message_box.critical_args,
            ((msg, self.controller.view), {}))


class ForgottenPasswordControllerTestCase(BaseTestCase):
    """Test the controller of the fogotten password page."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup the tests."""
        yield super(ForgottenPasswordControllerTestCase, self).setUp()
        self.view = FakeForgottenPasswordPageView()
        self.backend = self.view
        self.controller = ForgottenPasswordController(
            message_box=FakeMessageBox())
        self.controller.view = self.view
        self.controller.backend = self.backend

    def test_register_fields(self):
        """Ensure that all the diff fields are registered."""
        self.controller._register_fields()
        self.assertEqual(self.view.field('email_address'),
            self.view.email_address_line_edit)

    def test_set_translated_strings(self):
        """Ensure that the correct strings are translated."""
        self.controller._set_translated_strings()
        self.assertEqual(self.view.forgotted_password_intro_label.text(),
            REQUEST_PASSWORD_TOKEN_LABEL % {'app_name':
                                            self.view.fake_wizard.app_name})
        self.assertEqual(self.view.email_address_label.text(),
            EMAIL_LABEL)
        self.assertEqual(self.view.send_button.text(),
            RESET_PASSWORD)
        self.assertEqual(self.view.try_again_button.text(),
            TRY_AGAIN_BUTTON)

    def test_set_enhanced_line_edit(self):
        """Test that the correct line enhancements have been added."""
        self.controller._set_enhanced_line_edit()
        elements = self.view.properties['validation_rule']
        self.assertFalse(elements is None)
        for e in elements:
            self.assertTrue(e[0] is self.view.email_address_line_edit)
            self.assertIsInstance(e[1], collections.Callable)

    def test_connect_ui(self):
        """Test that the correct ui signals are connected."""
        self.controller._connect_ui()
        self.assertEqual(self.view.email_address_line_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)
        self.assertIsInstance(self.view.send_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.try_again_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.on_password_reset_token_sent_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_password_reset_error_cb,
            collections.Callable)

    def test_on_try_again(self):
        """Test that the on_try_again callback does work as expected."""
        self.controller.on_try_again()
        self.assertFalse(self.view.try_again_widget.isVisible())
        self.assertTrue(self.view.email_widget.isVisible())

    def test_on_password_reset_token_sent(self):
        """Test that the on_password_token_sent callback works as expected."""
        self.controller.on_password_reset_token_sent()
        self.assertTrue(self.view.properties['wizard_next'])
        self.assertEqual(self.view.next,
            self.view.fake_wizard.reset_password_page_id)


class ResetPasswordControllerTestCase(BaseTestCase):
    """Ensure that the reset password works as expected."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup the tests."""
        yield super(ResetPasswordControllerTestCase, self).setUp()
        self.view = ResetPasswordPageView()
        self.backend = self.view
        self.controller = ResetPasswordController()
        self.controller.view = self.view
        self.controller.backend = self.backend

    def test_set_translated_strings(self):
        """Ensure that the correct strings are set."""
        self.controller._set_translated_strings()
        self.assertEqual(self.view.reset_password_button.text(),
            RESET_PASSWORD)
        self.assertEqual(self.view.properties['subtitle'],
            PASSWORD_HELP)

    def test_connect_ui(self):
        """Ensure that the diffent signals from the ui are connected."""
        self.controller._connect_ui()
        self.assertIsInstance(self.view.reset_password_button.function,
            collections.Callable)
        self.assertIsInstance(self.view.on_password_changed_cb,
            collections.Callable)
        self.assertIsInstance(self.view.on_password_change_error_cb,
            collections.Callable)
        self.assertEqual(self.view.reset_code_line_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)
        self.assertEqual(self.view.password_line_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)
        self.assertEqual(self.view.confirm_password_line_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)

    def test_add_line_edits_validations(self):
        """Ensure that the line validation have been added."""
        self.controller._add_line_edits_validations()
        self.assertEqual(self.view.password_line_edit.receivers(
            SIGNAL('textChanged(QString)')), 1)
        elements = self.view.properties['validation_rule']
        self.assertFalse(elements is None)
        for e in elements:
            self.assertTrue(type(e[0]) is QLineEdit)
            self.assertIsInstance(e[1], collections.Callable)

    def test_set_new_password(self):
        """Test that the correct action is performed."""
        email = 'email@email.com'
        code = 'qwe123'
        password = 'T3st3rqw'
        self.view.wizard().forgotten.ui.email_line_edit.setText(email)
        self.view.reset_code_line_edit.setText(code)
        self.view.password_line_edit.setText(password)
        self.controller.set_new_password()
        self.assertEqual(self.view.properties['backend_new_password'],
            ((self.view.fake_wizard.app_name, email, code, password), {}))

    def test_is_correct_password_confirmation_true(self):
        """Test that the correct password confirmation is used."""
        password = 'password'
        self.view.password_line_edit.setText(password)
        self.assertTrue(self.controller.is_correct_password_confirmation(
                                                                    password))

    def test_is_correct_password_confirmation_false(self):
        """Test that the correct password confirmation is used."""
        password = 'password'
        self.view.password_line_edit.setText(password + password)
        self.assertFalse(self.controller.is_correct_password_confirmation(
                                                                    password))


class ResetPasswordControllerValidationTest(BaseTestCase):

    """Tests for ResetPasswordController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(ResetPasswordControllerValidationTest, self).setUp()
        self.controller = ResetPasswordController()
        self.controller.view = FakeResetPasswordPage()

    def test_page_initialized(self):
        """Test the initial state of the page when it is loaded."""
        self.controller.pageInitialized()
        self.assertFalse(
            self.controller.view.reset_password_button.enabled())
        self.assertTrue(
            self.controller.view.reset_password_button.properties['default'])
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.reset_password_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_valid(self):
        """Enable the button with valid data."""
        self.controller.view.reset_code_line_edit.setText("ABCD")
        self.controller.view.password_line_edit.setText("1234567A")
        self.controller.view.confirm_password_line_edit.setText("1234567A")
        self.controller._validate()
        self.assertTrue(self.controller.view.reset_password_button.enabled())
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.reset_password_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_invalid_code(self):
        """Disable the button with an invalid code."""
        self.controller.view.reset_code_line_edit.setText("")
        self.controller.view.password_line_edit.setText("1234567A")
        self.controller.view.confirm_password_line_edit.setText("1234567A")
        self.controller._validate()
        self.assertFalse(self.controller.view.reset_password_button.enabled())
        self.assertTrue("DisabledState" in self.controller.view.properties)
        self.assertEqual(self.controller.view.properties["DisabledState"],
            not self.controller.view.reset_password_button.enabled())
        self.assertTrue(self.controller.view.properties.get('unpolish', False))
        self.assertTrue(self.controller.view.properties.get('polish', False))

    def test_invalid_password(self):
        """Disable the button with an invalid password."""
        self.controller.view.reset_code_line_edit.setText("")
        self.controller.view.password_line_edit.setText("1234567")
        self.controller.view.confirm_password_line_edit.setText("1234567")
        self.controller._validate()
        self.assertFalse(self.controller.view.reset_password_button.enabled())

    def test_invalid_confirm(self):
        """Disable the button with an invalid password confirm."""
        self.controller.view.reset_code_line_edit.setText("")
        self.controller.view.password_line_edit.setText("1234567A")
        self.controller.view.confirm_password_line_edit.setText("1234567")
        self.controller._validate()
        self.assertFalse(self.controller.view.reset_password_button.enabled())


class ResetPasswordControllerRealControllerTest(BaseTestCase):

    """Tests for ResetPasswordController, but without Mocker."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup test."""
        yield super(ResetPasswordControllerRealControllerTest, self).setUp()
        self.controller = ResetPasswordController()
        self.controller.view = FakeWizardForResetPassword()

    def test_on_password_changed(self):
        """Test that on_password_changed execute the proper operation."""
        self.controller.on_password_changed('app_name', '')
        self.assertTrue(self.controller.view.hide_value)
        times_visited = 2
        self.assertEqual(self.controller.view.count_back, times_visited)
        self.assertEqual(self.controller.view.text_value, 'mail@mail.com')

    def test_on_password_changed_not_visited(self):
        """Test that on_password_changed execute the proper operation."""
        current_user_page_id = 20
        self.patch(self.controller.view, "current_user_page_id",
            current_user_page_id)
        self.controller.on_password_changed('app_name', '')
        self.assertTrue(self.controller.view.hide_value)
        times_visited = 4
        self.assertEqual(self.controller.view.count_back, times_visited)
        self.assertEqual(self.controller.view.text_value, 'mail@mail.com')
